/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 2000-2007 Sun Microsystems, Inc. All rights reserved. 
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License ("CDDL") (collectively, the "License").  You may
 * not use this file except in compliance with the License.  You can obtain
 * a copy of the License at https://glassfish.dev.java.net/public/CDDL+GPL.html
 * or mq/legal/LICENSE.txt.  See the License for the specific language
 * governing permissions and limitations under the License.
 * 
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at mq/legal/LICENSE.txt.  Sun designates
 * this particular file as subject to the "Classpath" exception as provided by
 * Sun in the GPL Version 2 section of the License file that accompanied this
 * code.  If applicable, add the following below the License Header, with the
 * fields enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 * 
 * Contributor(s):
 * 
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or  to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright holder. 
 */

/*
 * @(#)JMSBroker.java	1.6 06/29/07
 */ 

package com.sun.messaging.jmq.jmsservice;

import java.util.Properties;

/**
 *
 */
public interface JMSBroker {
    
    /**
     *  Parse broker command line and convert the args into a hashtable format.<p>
     *
     *  Additional arguments are:
     *  <UL>
     *      <LI> -varhome: The location of the VAR directory to use</LI>
     *      <LI> -imqhome: The location of the base IMQ directory</LI>
     *  </UL>
     *
     *  @param args The broker arguments in broker command line format.
     *
     *  @return The resulting Properties that represent the command line
     *          parameters passed in.
     *
     *  @throws IllegalArguementException   If args contain any invalid option.
     */
    public Properties parseArgs(String[] args)
    throws IllegalArgumentException;

    /**
     *  Initialize the broker. Only one broker can be running in a given JVM.
     *  The call returns as soon as the broker sucessfully initializes. The
     *  exact same paramters that are passed to initialize need to be passed to
     *  {@code start{}}
     *
     *  @param  inProcess   Indicates that the broker is running inprocess
     *                      and the shutdown hook and memory management
     *                      code should not be used.
     *  @param  properties  The configuration properties for the broker.
     *
     *  @param  el          An optional class to notify when a broker has
     *                      completed starting or has been shutdown.
     *
     *  @return The exit code returned by the broker. This is the same value
     *          returned as would be if the broker was running as a standalone
     *          process.<br>
     *          {@code 0} - if it was started successfully<br>.
     *          {@code non-zero} - otherwise
     *
     *  @throws OutOfMemoryError    If the broker can not allocate enough
     *                              memory to continue running.
     *  @throws IllegalStateException   If the broker is already running.  
     *  @throws IllegalArgumentException    If an invalid value for a property
     *                                      was passed in {@code properties}.
     */
    public int init(boolean inProcess,
            Properties properties, BrokerEventListener el)
    throws OutOfMemoryError, IllegalStateException, IllegalArgumentException;
    
    
    /**
     *  Start the broker. Only one broker can be running in a single JVM.
     *  The call returns as soon as the broker successfully starts.
     *
     *  @param  inProcess   indicates that the broker is running inprocess
     *                      and the shutdown hook and memory management
     *                      code should not be used.
     *
     *  @param  properties  the configuration properties for the broker.
     *
     *  @param  el          An optional class to notify when a broker has
     *                      completed starting or has been shutdown.
     *
     *  @return The exit code returned by the broker. This is the same value
     *          returned as would be if the broker was running as a standalone
     *          process.<br>
     *          {@code 0} - if it was started successfully<br>.
     *          {@code non-zero} - otherwise
     *
     *  @throws OutOfMemoryError    If the broker can not allocate enough
     *                              memory to continue running.
     *  @throws IllegalStateException   If the broker is already running.  
     *  @throws IllegalArgumentException    If an invalid value for a property
     *                                      was passed in {@code properties}.
     */
    public int start(boolean inProcess, 
            Properties properties, BrokerEventListener el)
    throws OutOfMemoryError, IllegalStateException, IllegalArgumentException;
    
    /**
     *  Stop the broker. Only one broker can be running in a single JVM
     *
     *  @param  cleanup {@code false} indicates that the broker does not have
     *                  to clean up; free resources etc. since the it is about
     *                  to exit.<br>
     *                  {@code true} indicates that the broker does have to
     *                  clean up/free resources etc.
     *
     *  @throws IllegalStateException if the broker is already stopped.  
     */
    public void stop(boolean cleanup)
    throws IllegalStateException;
    
    /**
     *  Return the default JMS Service that supports 'DIRECT' in-JVM Java EE JMS
     *  clients.
     *
     *  @throws IllegalStateException if the broker is already stopped
     * 
     */
    public JMSService getJMSService()
    throws IllegalStateException;

    /**
     *  Return the named JMS Service that supports 'DIRECT' in-JVM Java EEJMS
     *  clients.
     *
     *  @param  serviceName The name of the service to return
     *
     *  @throws IllegalStateException if the broker is already stopped
     */
    public JMSService getJMSService(String serviceName)
    throws IllegalStateException;
    
}
